#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2008-2011 Richard Dymond (rjdymond@gmail.com)
#
# This file is part of SkoolKit.
#
# SkoolKit is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# SkoolKit is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# SkoolKit. If not, see <http://www.gnu.org/licenses/>.

import sys
import os
from os.path import isfile, isdir, basename, dirname
import posixpath
import shutil
import time
from skoolkit import version
from skoolkit.skoollib import get_skool_class
from skoolkit.skoolhtml import get_skool_writer_class, join, FileInfo, SkoolWriter, NULL_IMG_LIB, gd, Image
from skoolkit.skoolparser import SkoolParser, RefParser, CASE_UPPER, CASE_LOWER, BASE_10, BASE_16

verbose = True

SEARCH_DIRS = ('', 'src', '/usr/share/skoolkit')
ALL_FILES = 'BbdGghiLmoPpty'

def error(msg):
    sys.stderr.write('ERROR: %s\n' % msg)
    sys.exit(1)

def notify(notice):
    if verbose:
        print notice

def clock(operation, prefix, *args, **kwargs):
    if verbose:
        sys.stdout.write('%s ' % prefix)
        sys.stdout.flush()
        go = time.time()
    result = operation(*args, **kwargs)
    if verbose:
        sys.stdout.write('(%0.2fs)\n' % (time.time() - go))
    return result

def find(fname, search_dir=None):
    if fname is None:
        return
    search_dirs = [search_dir] if search_dir else []
    search_dirs.extend(SEARCH_DIRS)
    for f in [join(d, fname) for d in search_dirs]:
        if isfile(f):
            return f

def parse_args(args):
    verbose = True
    new_images = False
    img_lib = None
    case = None
    base = None
    infiles = []
    files = ''
    pages = []
    output_dir = None
    i = 0
    while i < len(args):
        arg = args[i]
        if arg == '-V':
            sys.stdout.write('SkoolKit %s\n' % version)
            sys.exit(0)
        if arg == '-q':
            verbose = False
        elif arg == '-d':
            output_dir = args[i + 1]
            i += 1
        elif arg == '-o':
            new_images = True
        elif arg == '-i':
            img_lib = args[i + 1]
            i += 1
        elif arg == '-l':
            case = CASE_LOWER
        elif arg == '-u':
            case = CASE_UPPER
        elif arg == '-D':
            base = BASE_10
        elif arg == '-H':
            base = BASE_16
        elif arg == '-P':
            pages += args[i + 1].split(',')
            i += 1
        elif arg[:2] == '-w':
            files += arg[2:]
        elif len(arg) > 1 and arg[0] == '-':
            print_usage()
        else:
            infiles.append(arg)
        i += 1

    if not infiles:
        print_usage()

    return verbose, new_images, img_lib, case, base, infiles, files, pages, output_dir

def print_usage():
    sys.stderr.write("""Usage: %s [options] FILE [FILE...]

  Convert skool files and ref files to HTML. FILE may be a regular file, or '-'
  for standard input.

Options:
  -V       Show SkoolKit version number and exit
  -q       Be quiet
  -d DIR   Write files in this directory
  -o       Overwrite existing image files
  -i LIB   Use this image library ('gd' or 'PIL'; default is 'gd')
  -l       Write disassembly in lower case
  -u       Write disassembly in upper case
  -D       Write disassembly in decimal
  -H       Write disassembly in hexadecimal
  -P LIST  Write only these custom pages (when '-wP' is specified)
  -wX      Write only these files, where X is one or more of:
             B = Graphic glitches
             b = Bugs
             d = Disassembly files
             G = Game status buffer
             g = Graphics
             i = Disassembly index
             m = Memory maps
             o = Other code
             P = Pages defined in the ref file
             p = Pokes
             t = Trivia
             y = Glossary
""" % basename(sys.argv[0]))
    sys.exit(1)

def process_file(infile, topdir, files, case, base, pages, img_lib, new_images):
    if infile == '-' or infile.endswith('.skool'):
        reffile = None
        skoolfile = infile
    else:
        reffile = infile
        skoolfile = None

    ref_parser = RefParser()
    reffile_f = find(reffile)
    config = {}
    search_dir = None

    if reffile:
        if reffile_f is None:
            error('%s: ref file not found' % reffile)
        search_dir = dirname(reffile_f)
        ref_parser.parse(reffile_f)
        config = ref_parser.build_dictionary('Config')
        prefix = basename(reffile)
        if prefix[-4:] == '.ref':
            prefix = prefix[:-4]
        skoolfile = config.get('SkoolFile', '%s.skool' % prefix)

    skoolfile_f = skoolfile if skoolfile == '-' else find(skoolfile, search_dir)
    if skoolfile_f is None:
        error('%s: skool file not found' % skoolfile)

    prefix = basename(skoolfile_f)
    if prefix == '-':
        prefix = 'program'
    elif prefix[-6:] == '.skool':
        prefix = prefix[:-6]

    if reffile is None and skoolfile != '-':
        search_dir = dirname(skoolfile_f)
        reffile_f = find('%s.ref' % prefix, search_dir)
        if reffile_f:
            ref_parser.parse(reffile_f)
            config = ref_parser.build_dictionary('Config')

    if skoolfile and skoolfile != '-':
        if reffile_f:
            notify('Using ref file: %s' % reffile_f)
        else:
            notify('Found no ref file for %s' % skoolfile_f)
    elif reffile:
        notify('Using skool file: %s' % skoolfile_f)

    skool_class = get_skool_class(config.get('SkoolClass'))
    skool_writer_class = get_skool_writer_class(config.get('SkoolWriterClass'))
    game_dir = config.get('GameDir', prefix)

    # Parse the skool file and initialise the writer
    fname = 'standard input' if skoolfile_f == '-' else skoolfile_f
    skool_parser = clock(SkoolParser, 'Parsing %s' % fname, skoolfile_f, case=case, base=base, html=True)
    skool = skool_class(skool_parser, ref_parser)
    file_info = FileInfo(topdir, game_dir, new_images)
    skool_writer = skool_writer_class(skool, file_info, case, img_lib)
    paths = skool_writer.paths

    # Check that the specified pages exist
    all_page_ids = skool_writer.get_page_ids()
    for page_id in pages:
        if page_id not in all_page_ids:
            error('Invalid page ID: %s' % page_id)

    # Create the disassembly subdirectory if necessary
    odir = file_info.odir
    if not isdir(odir):
        notify('Creating directory %s' % odir)
        os.makedirs(odir)

    # Copy style.css and skool.js if necessary
    for fname, dest in ((paths.get('StyleSheet'), paths.get('StyleSheetPath', '')), (paths.get('JavaScript'), paths.get('JavaScriptPath', ''))):
        if fname:
            f = find(fname, search_dir)
            if f:
                html_f = posixpath.normpath(join(topdir, game_dir, dest, fname))
                if not isfile(html_f) or os.stat(f).st_mtime > os.stat(html_f).st_mtime:
                    notify('Copying %s to %s' % (f, html_f))
                    shutil.copy2(f, html_f)
            else:
                error('<%s>: file not found' % fname)

    # Write logo image file if necessary
    if skool_writer.need_image(paths['Logo']) and skool_writer.write_logo_image(odir):
        notify('  Wrote %s' % join(game_dir, paths['Logo']))

    # Write disassembly files
    if 'd' in files:
        clock(skool_writer.write_asm_entries, '  Writing disassembly files in %s' % join(game_dir, skool.code_path))

    # Write the memory map files
    if 'm' in files:
        clock(skool_writer.write_memory_map, '  Writing %s' % join(game_dir, paths['MemoryMap']))
        clock(skool_writer.write_routines_map, '  Writing %s' % join(game_dir, paths['RoutinesMap']))
        clock(skool_writer.write_data_map, '  Writing %s' % join(game_dir, paths['DataMap']))
        clock(skool_writer.write_messages_map, '  Writing %s' % join(game_dir, paths['MessagesMap']))

    # Write pages defined in the ref file
    if 'P' in files:
        for page_id in pages or all_page_ids:
            clock(skool_writer.write_page, '  Writing %s' % join(game_dir, paths[page_id]), page_id)

    # Write other files
    if 'G' in files and skool_writer.has_gbuffer():
        clock(skool_writer.write_gbuffer, '  Writing %s' % join(game_dir, paths['GameStatusBuffer']))
    if 'g' in files and skool_writer.has_graphics():
        clock(skool_writer.write_graphics, '  Writing %s' % join(game_dir, paths['Graphics']))
    if 'B' in files and skool_writer.has_graphic_glitches():
        clock(skool_writer.write_graphic_glitches, '  Writing %s' % join(game_dir, paths['GraphicGlitches']))
    if 'b' in files and skool_writer.has_bugs():
        clock(skool_writer.write_bugs, '  Writing %s' % join(game_dir, paths['Bugs']))
    if 't' in files and skool_writer.has_facts():
        clock(skool_writer.write_facts, '  Writing %s' % join(game_dir, paths['Facts']))
    if 'y' in files and skool_writer.has_glossary():
        clock(skool_writer.write_glossary, '  Writing %s' % join(game_dir, paths['Glossary']))
    if 'p' in files and skool_writer.has_pokes():
        clock(skool_writer.write_pokes, '  Writing %s' % join(game_dir, paths['Pokes']))
    if 'o' in files:
        for code_id, code in skool.other_code:
            skoolfile = find(code['Source'], search_dir)
            skool2_parser = clock(SkoolParser, '  Parsing %s' % skoolfile, skoolfile, case=case, base=base, html=True)
            skool2 = skool_class(skool2_parser, ref_parser, code_id)
            skool2_writer = skool_writer_class(skool2, file_info, case, img_lib)
            path = code['Path']
            index = code['Index']
            clock(skool2_writer.write_map, '    Writing %s' % join(game_dir, index), index, code['Title'], '', path)
            clock(skool2_writer.write_entries, '    Writing disassembly files in %s' % join(game_dir, path), path, index, code['Header'])

    # Write index.html
    if 'i' in files:
        clock(skool_writer.write_index, '  Writing %s' % join(game_dir, paths['GameIndex']))

def run(new_images, img_lib, case, base, infiles, files, pages, output_dir):
    if img_lib == NULL_IMG_LIB:
        img_module = img_lib
    elif img_lib == 'PIL':
        img_module = Image or gd
    else:
        img_module = gd or Image
    if img_module:
        if img_module == NULL_IMG_LIB:
            actual_img_lib = img_module
        elif img_module is Image:
            actual_img_lib = 'PIL'
        else:
            actual_img_lib = 'gd'
        if img_lib:
            notify("Using '%s' image library" % actual_img_lib)
    else:
        actual_img_lib = None
        sys.stderr.write('WARNING: No image library found; no images will be created\n')

    if output_dir in (None, '.'):
        topdir = ''
    else:
        pdir, bdir = os.path.split(output_dir)
        if not bdir:
            pdir, bdir = os.path.split(pdir)
        topdir = join(pdir, bdir)
        if not isdir(topdir):
            # Create the top-level directory
            notify('Creating directory %s' % topdir)
            os.makedirs(topdir)

    for infile in infiles:
        process_file(infile, topdir, files, case, base, pages, actual_img_lib, new_images)

def main():
    global verbose
    verbose, new_images, img_lib, case, base, infiles, files, pages, output_dir = parse_args(sys.argv[1:])
    run(new_images, img_lib, case, base, infiles, files or ALL_FILES, pages, output_dir)

if __name__ == "__main__":
    main()
